###############################################################################
#
# Copyright 2014 by Shoobx, Inc.
#
###############################################################################


class MigrantException(RuntimeError):
    pass


class ConfigurationError(MigrantException):
    pass


class ScriptNotFoundError(MigrantException):
    def __str__(self):
        return "Script not found: %s" % self.args


class ScriptAlreadyExists(MigrantException):
    def __str__(self):
        return "Script already exists: %s" % self.args


class RepositoryNotFound(MigrantException):
    def __str__(self):
        return "Repository not found: %s" % self.args


class BackendNotRegistered(MigrantException):
    def __str__(self):
        return "Backend not registered: %s" % self.args


class BackendNameConflict(MigrantException):
    def __str__(self):
        return "Backend name conflict in these pacakges: %s" % self.args


class DatabaseUnavailable(MigrantException):
    """Raised by backend when database is not available for processing"""



###############################################################################
#
# Copyright 2014 by Shoobx, Inc.
#
###############################################################################
import os
import io
import unittest
import tempfile
import shutil
import textwrap
import logging
import pytest
from configparser import ConfigParser
import multiprocessing

from migrant import cli, backend, exceptions


HERE = os.path.dirname(__file__)

SAMPLE_CONFIG = """
[db1]
backend = mongo
repository = repo/
backend_uri = localhost:27017/acme

[db2]
backend = test
repository = /repo
"""

INTEGRATION_CONFIG = """
[test]
backend = test
repository = %s
""" % os.path.join(
    HERE, "scripts"
)

INTEGRATION_CONFIG += """
[virgin]
backend = test
repository = %s
""" % os.path.join(
    HERE, "noscripts"
)


class MockedDb:
    def __init__(self, name, manager):
        self.name = name
        self.migrations = manager.list()
        self.data = manager.dict()


class MockedBackend(backend.MigrantBackend):
    def __init__(self, dbs, manager=None):
        self.dbs = dbs
        self.new_scripts = manager.list() if manager else []
        self.inits = 0

    def begin(self, db):
        return db

    def list_migrations(self, db):
        return list(db.migrations)

    def push_migration(self, db, migration):
        db.migrations.append(migration)

    def pop_migration(self, db, migration):
        db.migrations.remove(migration)

    def generate_connections(self):
        """Generate connections to process
        """
        yield from self.dbs

    def generate_test_connections(self):
        return self.generate_connections()

    def on_new_script(self, rev_name):
        """Called when new script is created
        """
        self.new_scripts.append(rev_name)

    def on_repo_init(self):
        """Called when new script repository is initialized
        """
        self.inits += 1


class ConfigTest(unittest.TestCase):
    def test_get_db_config(self):
        cp = ConfigParser()
        cp.read_file(io.StringIO(SAMPLE_CONFIG), "SAMPLE_CONFIG")
        config = cli.get_db_config(cp, "db1")
        self.assertEqual(
            config,
            {
                "backend": "mongo",
                "backend_uri": "localhost:27017/acme",
                "repository": "repo/",
            },
        )


class UpgradeTest(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def backend_fixture(self, tmpdir, migrant_backend):
        m = multiprocessing.Manager()
        self.db0 = MockedDb("db0", m)
        self.backend = MockedBackend([self.db0], m)
        migrant_backend.set(self.backend)

        migrant_ini = tmpdir.join("migrant.ini")
        migrant_ini.write(INTEGRATION_CONFIG)
        self.migrant_ini = str(migrant_ini)

        self.cfg = cli.load_config(self.migrant_ini)

        self.logstream = io.StringIO()
        hdl = logging.StreamHandler(self.logstream)
        hdl.setFormatter(logging.Formatter(logging.BASIC_FORMAT))
        logging.root.addHandler(hdl)
        logging.root.setLevel(logging.INFO)

    def test_status_dirty(self):
        self.db0.migrations.extend(["aaaa_first"])

        args = cli.parser.parse_args(["test", "status"])
        cli.dispatch(args, self.cfg)

        log = self.logstream.getvalue()
        self.assertIn("Pending actions: 2", log)

    def test_status_clean(self):
        self.db0.migrations.extend(["aaaa_first", "bbbb_second", "cccc_third"])

        args = cli.parser.parse_args(["test", "status"])
        cli.dispatch(args, self.cfg)

        log = self.logstream.getvalue()
        self.assertIn("Up-to-date", log)

    def test_no_scripts(self):
        args = cli.parser.parse_args(["virgin", "upgrade"])
        cli.dispatch(args, self.cfg)
        self.assertEqual(list(self.db0.migrations), ["INITIAL"])

    def test_initial_upgrade(self):
        args = cli.parser.parse_args(["test", "upgrade"])
        cli.dispatch(args, self.cfg)

        self.assertEqual(
            list(self.db0.migrations),
            ["INITIAL", "aaaa_first", "bbbb_second", "cccc_third"],
        )

    def test_subsequent_emtpy_upgrade(self):
        args = cli.parser.parse_args(["test", "upgrade"])
        cli.dispatch(args, self.cfg)
        # this should be a noop
        cli.dispatch(args, self.cfg)

        self.assertEqual(
            list(self.db0.migrations),
            ["INITIAL", "aaaa_first", "bbbb_second", "cccc_third"],
        )

    def test_upgrade_latest(self):
        self.db0.migrations.extend(["aaaa_first"])

        args = cli.parser.parse_args(["test", "upgrade"])
        cli.dispatch(args, self.cfg)

        self.assertEqual(
            list(self.db0.migrations), ["aaaa_first", "bbbb_second", "cccc_third"]
        )
        self.assertEqual(dict(self.db0.data), {"hello": "world", "value": "c"})

    def test_upgrade_particular(self):
        self.db0.migrations.extend(["aaaa_first"])
        args = cli.parser.parse_args(["test", "upgrade", "--revision", "bbbb_second"])
        cli.dispatch(args, self.cfg)

        self.assertEqual(list(self.db0.migrations), ["aaaa_first", "bbbb_second"])
        self.assertEqual(dict(self.db0.data), {"value": "b"})

    def test_downgrade(self):
        self.db0.migrations.extend(
            ["INITIAL", "aaaa_first", "bbbb_second", "cccc_third"]
        )
        self.db0.data.update({"hello": "world", "value": "c"})

        args = cli.parser.parse_args(["test", "upgrade", "--revision", "aaaa_first"])
        cli.dispatch(args, self.cfg)

        self.assertEqual(list(self.db0.migrations), ["INITIAL", "aaaa_first"])
        self.assertEqual(dict(self.db0.data), {"value": "a"})

    def test_downgrade_to_initial(self):
        self.db0.migrations.extend(
            ["INITIAL", "aaaa_first", "bbbb_second", "cccc_third"]
        )
        self.db0.data.update({"hello": "world", "value": "c"})

        args = cli.parser.parse_args(["test", "upgrade", "--revision", "INITIAL"])
        cli.dispatch(args, self.cfg)

        self.assertEqual(list(self.db0.migrations), ["INITIAL"])
        self.assertEqual(dict(self.db0.data), {})

    def test_dry_run_upgrade(self):
        self.db0.migrations.extend(["aaaa_first"])
        self.db0.data.update({"value": "a"})

        args = cli.parser.parse_args(["test", "upgrade", "--dry-run"])
        cli.dispatch(args, self.cfg)

        self.assertEqual(list(self.db0.migrations), ["aaaa_first"])
        self.assertEqual(dict(self.db0.data), {"value": "a"})

    def test_test(self):
        self.db0.migrations = ["INITIAL", "aaaa_first", "bbbb_second", "cccc_third"]
        self.db0.data = {"hello": "world", "value": "c"}

        cli.main(["-c", self.migrant_ini, "test", "test"])

        log = self.logstream.getvalue()
        assert "Testing upgrade" in log
        assert "Testing downgrade" in log
        assert "Testing completed" in log


class InitTest(unittest.TestCase):
    def setUp(self):
        self.dir = tempfile.mkdtemp("migrant")

    def tearDown(self):
        shutil.rmtree(self.dir)


@pytest.fixture
def sample_config(tmpdir):
    SAMPLE_CONFIG = textwrap.dedent(
        """
    [newdb]
    backend = noop
    repository = %s/repo
    """
        % tmpdir
    )

    cfg = ConfigParser()
    cfg.read_file(io.StringIO(SAMPLE_CONFIG), "SAMPLE_CONFIG")
    return cfg


def initialize(cfg):
    # Initialize repository
    args = cli.parser.parse_args(["newdb", "init"])
    cli.dispatch(args, cfg)


def get_scripts_filename(cfg):
    repodir = cfg.get("newdb", "repository")
    return os.path.join(repodir, "scripts.lst")


def test_missing_cmd():
    with pytest.raises(SystemExit):
        cli.parser.parse_args(["newdb"])


def test_init(sample_config):
    args = cli.parser.parse_args(["newdb", "init"])
    cli.dispatch(args, sample_config)

    repodir = sample_config.get("newdb", "repository")
    assert os.path.exists(repodir)


def test_init_subsequent(sample_config):
    os.mkdir(sample_config.get("newdb", "repository"))

    slistfname = get_scripts_filename(sample_config)
    with open(slistfname, "w") as f:
        f.write("aaaa_test.py")

    args = cli.parser.parse_args(["newdb", "init"])
    cli.dispatch(args, sample_config)

    with open(slistfname) as slist:
        assert slist.read() == "aaaa_test.py"


def test_new_uninitialized(sample_config):
    args = cli.parser.parse_args(["newdb", "new", "First script"])
    with pytest.raises(exceptions.RepositoryNotFound):
        cli.dispatch(args, sample_config)


def test_new_initialized(sample_config):
    initialize(sample_config)
    args = cli.parser.parse_args(["newdb", "new", "First script"])
    cli.dispatch(args, sample_config)

    with open(get_scripts_filename(sample_config)) as slist:
        lines = slist.readlines()
        assert len(lines) == 3
        assert lines[-1].strip().endswith("_first_script.py")


def test_new_subsequent(sample_config):
    initialize(sample_config)
    args = cli.parser.parse_args(["newdb", "new", "First script"])
    cli.dispatch(args, sample_config)

    args = cli.parser.parse_args(["newdb", "new", "Second script"])
    cli.dispatch(args, sample_config)

    with open(get_scripts_filename(sample_config)) as slist:
        lines = slist.readlines()
        assert len(lines) == 4
        assert lines[-2].strip().endswith("_first_script.py")
        assert lines[-1].strip().endswith("_second_script.py")


def test_new_duplicate(sample_config):
    initialize(sample_config)
    args = cli.parser.parse_args(["newdb", "new", "First script"])
    cli.dispatch(args, sample_config)

    with pytest.raises(exceptions.ScriptAlreadyExists):
        cli.dispatch(args, sample_config)


def test_new_on_repo_init(sample_config, migrant_backend):
    backend = MockedBackend([])
    migrant_backend.set(backend)

    args = cli.parser.parse_args(["newdb", "init"])
    cli.dispatch(args, sample_config)

    assert backend.inits == 1


def test_new_on_new_script(sample_config, migrant_backend):
    backend = MockedBackend([])
    migrant_backend.set(backend)
    initialize(sample_config)

    args = cli.parser.parse_args(["newdb", "new", "First script"])
    cli.dispatch(args, sample_config)

    assert len(backend.new_scripts) == 1
    assert backend.new_scripts[0].endswith("_first_script")



###############################################################################
#
# Copyright 2014 by Shoobx, Inc.
#
###############################################################################
from typing import List, Dict, Generator
import os
import unittest
import time

import mock

from migrant import exceptions
from migrant.engine import MigrantEngine
from migrant.backend import MigrantBackend
from migrant.repository import Script, Repository


class MigrantEngineTest(unittest.TestCase):
    def test_calc_actions_simple(self):
        engine = _make_engine(["a", "b"], ["a", "b", "c", "d"])
        actions = engine.calc_actions(None, "d")
        self.assertEqual(actions, [("+", "c"), ("+", "d")])

    def test_calc_actions_downgrade(self):
        engine = _make_engine(["a", "b", "c"], ["a", "b", "c", "d"])
        actions = engine.calc_actions(None, "a")
        self.assertEqual(actions, [("-", "c"), ("-", "b")])

    def test_calc_actions_outoforder_upgrade(self):
        engine = _make_engine(["a", "b", "d"], ["a", "b", "c", "d"])
        actions = engine.calc_actions(None, "d")
        self.assertEqual(actions, [("+", "c")])

    def test_calc_actions_outoforder_downgrade(self):
        engine = _make_engine(["a", "b", "d"], ["a", "b", "c", "d"])
        actions = engine.calc_actions(None, "c")
        self.assertEqual(actions, [("-", "d"), ("+", "c")])

    def test_calc_actions_nobeginning(self):
        engine = _make_engine(["b", "d"], ["a", "b", "c", "d", "e"])
        actions = engine.calc_actions(None, "e")
        self.assertEqual(actions, [("+", "c"), ("+", "e")])

    def test_calc_actions_removedscripts(self):
        engine = _make_engine(["a", "b", "c", "d"], ["c", "d", "e"])
        actions = engine.calc_actions(None, "e")
        self.assertEqual(actions, [("+", "e")])

    def test_calc_actions_wrongorder(self):
        engine = _make_engine(["d", "b", "a"], ["b", "c", "d", "e"])
        actions = engine.calc_actions(None, "e")
        self.assertEqual(actions, [("+", "c"), ("+", "e")])

    def test_calc_actions_nocommon(self):
        engine = _make_engine(["a", "b", "c"], ["d", "e"])
        actions = engine.calc_actions(None, "e")
        self.assertEqual(actions, [("+", "d"), ("+", "e")])

    def test_calc_actions_wrongorder_downgrade(self):
        engine = _make_engine(["e", "d", "b", "a"], ["b", "c", "d", "e"])
        actions = engine.calc_actions(None, "b")
        self.assertEqual(actions, [("-", "e"), ("-", "d")])

    def test_revert_actions(self):
        engine = _make_engine([], [])
        reverted = engine.revert_actions([("-", "a"), ("+", "b")])
        self.assertEqual(reverted, [("-", "b"), ("+", "a")])

    def test_test(self):
        log = []
        engine = _make_engine(["a", "b"], ["a", "b", "c", "d"], log)
        engine.test()
        assert log == [
            # DB 1 PASS 1
            "db1 c before up",
            "db1 c up",
            "db1 c after up",
            "db1 d before up",
            "db1 d up",
            "db1 d after up",
            "db1 d before down",
            "db1 d down",
            "db1 d after down",
            "db1 c before down",
            "db1 c down",
            "db1 c after down",
            # DB 1 PASS 2
            "db1 c before up",
            "db1 c up",
            "db1 c after up",
            "db1 d before up",
            "db1 d up",
            "db1 d after up",
            "db1 d before down",
            "db1 d down",
            "db1 d after down",
            "db1 c before down",
            "db1 c down",
            "db1 c after down",
            # DB 2 PASS 1
            "db2 c before up",
            "db2 c up",
            "db2 c after up",
            "db2 d before up",
            "db2 d up",
            "db2 d after up",
            "db2 d before down",
            "db2 d down",
            "db2 d after down",
            "db2 c before down",
            "db2 c down",
            "db2 c after down",
            # DB 2 PASS 2
            "db2 c before up",
            "db2 c up",
            "db2 c after up",
            "db2 d before up",
            "db2 d up",
            "db2 d after up",
            "db2 d before down",
            "db2 d down",
            "db2 d after down",
            "db2 c before down",
            "db2 c down",
            "db2 c after down",
        ]


class ScriptMock:
    def __init__(self, name, log):
        self.name = name
        self.log = log

    def up(self, db):
        self.log.append(f"{db} {self.name} up")

    def down(self, db):
        self.log.append(f"{db} {self.name} down")

    def test_before_up(self, db):
        self.log.append(f"{db} {self.name} before up")

    def test_after_up(self, db):
        self.log.append(f"{db} {self.name} after up")

    def test_before_down(self, db):
        self.log.append(f"{db} {self.name} before down")

    def test_after_down(self, db):
        self.log.append(f"{db} {self.name} after down")


def _make_engine(migrations, scripts, log=None):
    """Make a mock engine that upgrades backend, having `migrations` installed,
    using `scripts` in repository
    """
    log = log if log is not None else []
    backend = mock.Mock()
    backend.list_migrations.return_value = migrations
    backend.begin = lambda db: db

    backend.generate_test_connections.return_value = ["db1", "db2"]

    scriptmodules = {sid: ScriptMock(sid, log) for sid in scripts}
    repository = mock.Mock()
    repository.list_script_ids.return_value = scripts
    repository.load_script = scriptmodules.__getitem__

    engine = MigrantEngine(backend, repository, {})
    return engine


class MultiDbBackend(MigrantBackend[str, str]):
    def __init__(self, dbs: List[str], logfname: str) -> None:
        self._applied = {}
        self.dbs = dbs
        self.logfname = logfname
        self.unavailable_dbs: List[str] = []
        for db in dbs:
            self._applied[db] = ["INITIAL"]

    def begin(self, db: str) -> str:
        if db in self.unavailable_dbs:
            raise exceptions.DatabaseUnavailable(db)
        return db

    def list_migrations(self, db: str) -> List[str]:
        return self._applied.get(db, [])

    def push_migration(self, db, migration):
        migrations = self._applied.setdefault(db, [])
        migrations.append(migration)

    def pop_migration(self, db, migration):
        migrations = self._applied.setdefault(db, [])
        migrations.remove(migration)

    def generate_connections(self) -> Generator[str, None, None]:
        """Generate connections to process
        """
        for db in self.dbs:
            yield db


class TimedScript(Script):
    def __init__(self, name: str, timemap: Dict[str, float], logfname: str) -> None:
        self.name = name
        self._timemap = timemap
        self._logfname = logfname

    def up(self, db):
        tosleep = self._timemap.get(db, 0)
        time.sleep(tosleep)
        with open(self._logfname, "a") as f:
            f.write(f"{db}: Upgraded to {self.name} ({tosleep}s)\n")


class MultiDbRepo(Repository):
    def __init__(self, timemap: Dict[str, float], logfname: str) -> None:
        self.timemap = timemap
        self.logfname = logfname

    def list_script_ids(self) -> List[str]:
        return ["script1"]

    def load_script(self, scriptid: str) -> Script:
        return TimedScript(scriptid, self.timemap, self.logfname)


def test_concurrent_upgrade_multiprocess(tmp_path) -> None:
    # GIVEN

    # Prepare the log file. Since we are running scripts in
    # multiprocessing environment, we cannot share memory data structure, like
    # list, and let scripts write log to it. Instead, write to a share file.
    logfname = os.path.join(tmp_path, "migration.log")
    # backend = MultiDbBackend(["db1"] + ["db2", "db3"]*10, logfname)
    backend = MultiDbBackend(["db1", "db2", "db3"], logfname)
    repository = MultiDbRepo({"db1": 0.1, "db2": 0.01, "db3": 0.05}, logfname)
    engine = MigrantEngine(backend, repository, {}, processes=2)

    # WHEN
    engine.update()

    # THEN
    with open(logfname, "r") as f:
        log = f.read().strip().split("\n")

    # Longest task executed last
    assert log == [
        "db2: Upgraded to script1 (0.01s)",
        "db3: Upgraded to script1 (0.05s)",
        "db1: Upgraded to script1 (0.1s)",
    ]


def test_concurrent_upgrade_singleprocess(tmp_path) -> None:
    # GIVEN

    # Prepare the log file. Since we are running scripts in
    # multiprocessing environment, we cannot share memory data structure, like
    # list, and let scripts write log to it. Instead, write to a share file.
    logfname = os.path.join(tmp_path, "migration.log")
    backend = MultiDbBackend(["db1", "db2", "db3"], logfname)
    repository = MultiDbRepo({"db1": 0.1, "db2": 0.01, "db3": 0.05}, logfname)
    engine = MigrantEngine(backend, repository, {}, processes=1)

    # WHEN
    engine.update()

    # THEN
    with open(logfname, "r") as f:
        log = f.read().strip().split("\n")

    # First task executed first despite being the longest
    assert log == [
        "db1: Upgraded to script1 (0.1s)",
        "db2: Upgraded to script1 (0.01s)",
        "db3: Upgraded to script1 (0.05s)",
    ]


def test_skip_unavailable(tmp_path) -> None:
    # GIVEN
    logfname = os.path.join(tmp_path, "migration.log")

    backend = MultiDbBackend(["db1", "db2", "db3"], logfname)
    backend.unavailable_dbs = ["db1", "db3"]
    repository = MultiDbRepo({"db1": 0.1, "db2": 0.01, "db3": 0.05}, logfname)
    engine = MigrantEngine(backend, repository, {}, processes=1)

    # WHEN
    engine.update()

    # THEN
    with open(logfname, "r") as f:
        log = f.read().strip().split("\n")

    # First task executed first despite being the longest
    assert log == [
        "db2: Upgraded to script1 (0.01s)",
    ]



###############################################################################
#
# Copyright 2014 by Shoobx, Inc.
#
###############################################################################
from typing import Dict
import os
import sys
import argparse
import logging
from configparser import ConfigParser

from migrant import exceptions
from migrant.engine import MigrantEngine
from migrant.backend import create_backend
from migrant.repository import create_repo

log = logging.getLogger(__name__)


def cmd_init(args, cfg):
    cfg = get_db_config(cfg, args.database)
    repo = create_repo(cfg)
    repo.init()
    backend = create_backend(cfg)
    backend.on_repo_init()


def cmd_new(args, cfg):
    cfg = get_db_config(cfg, args.database)
    repo = create_repo(cfg)
    revname = repo.new_script(args.title)
    backend = create_backend(cfg)
    backend.on_new_script(revname)


def cmd_upgrade(args, cfg: Dict[str, str]) -> None:
    cfg = get_db_config(cfg, args.database)
    repo = create_repo(cfg)
    backend = create_backend(cfg)
    engine = MigrantEngine(
        backend, repo, cfg, dry_run=args.dry_run, processes=args.parallel,
    )
    engine.update(args.revision)


def cmd_test(args, cfg):
    cfg = get_db_config(cfg, args.database)
    repo = create_repo(cfg)
    backend = create_backend(cfg)
    engine = MigrantEngine(backend, repo, cfg)
    engine.test(args.revision)


def cmd_status(args, cfg):
    cfg = get_db_config(cfg, args.database)
    repo = create_repo(cfg)
    backend = create_backend(cfg)
    engine = MigrantEngine(backend, repo, cfg)
    actions = engine.status()
    if actions:
        log.info("Pending actions: %s", actions)
    else:
        log.info("Up-to-date")


parser = argparse.ArgumentParser(description="Database Migration Engine")
parser.add_argument("database", help="Database name")

parser.add_argument(
    "-c", "--config", default="migrant.ini", help=("Config file to be used")
)

commands = parser.add_subparsers(dest="cmd")
commands.required = True

# INIT options
init_parser = commands.add_parser("init", help="Initialize migration script repository")
init_parser.set_defaults(cmd=cmd_init)
# init_parser.add_argument("database", help="Database name")

# NEW options
new_parser = commands.add_parser(
    "new", help="Create new migration script, add a script title argument!"
)
new_parser.set_defaults(cmd=cmd_new)
# new_parser.add_argument("database", help="Database name")
new_parser.add_argument("title", help="Migration script title")

# STATUS options
status_parser = commands.add_parser("status", help="Show the migration status")
status_parser.set_defaults(cmd=cmd_status)

# UPGRADE options
upgrade_parser = commands.add_parser("upgrade", help="Perform upgrade")
upgrade_parser.set_defaults(cmd=cmd_upgrade)
upgrade_parser.add_argument(
    "-n",
    "--dry-run",
    action="store_true",
    help=(
        "dry run: do not execute scripts, only " "show what is going to be executed."
    ),
)
# upgrade_parser.add_argument("database", help="Database name to upgrade")
upgrade_parser.add_argument(
    "-r",
    "--revision",
    help=("Revision to upgrade to. If not specified, " "latest revision will be used"),
)

upgrade_parser.add_argument(
    "-j",
    "--parallel",
    nargs="?",
    type=int,
    default=1,
    help=(
        "Migrate databases in parallel. If backend provides multiple databases, "
        "migration for each of them will be performed in parallel. Concurrency "
        "level is set by this argument."
    ),
)


# TEST options
test_parser = commands.add_parser(
    "test", help="Test pending migrations by going through update and downgrade"
)
test_parser.set_defaults(cmd=cmd_test)
test_parser.add_argument(
    "-r",
    "--revision",
    help=("Revision to upgrade to. If not specified, " "latest revision will be used"),
)


def load_config(fname):
    if not os.path.exists(fname):
        raise exceptions.ConfigurationError("%s is missing" % fname)
    cfg = ConfigParser()
    with open(fname) as cfgfp:
        cfg.read_file(cfgfp)
    return cfg


def get_db_config(cfg, name):
    if not cfg.has_section(name):
        ava = ", ".join(cfg.sections())
        raise exceptions.ConfigurationError(
            f"No database {name} in migrant.ini, available names: {ava}"
        )
    return dict(cfg.items(name))


def setup_logging(args, cfg):
    FORMAT = "%(asctime)s %(levelname)s %(module)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=FORMAT)


def dispatch(args, cfg):
    args.cmd(args, cfg)


def main(args=sys.argv[1:]):
    args = parser.parse_args(args)
    try:
        cfg = load_config(args.config)
        setup_logging(args, cfg)
        dispatch(args, cfg)
    except exceptions.MigrantException as e:
        sys.exit("fatal: %s" % e)



###############################################################################
#
# Copyright 2014 by Shoobx, Inc.
#
###############################################################################
from typing import Optional, TypeVar, Dict, List, Tuple, Generic
import logging
import multiprocessing
import functools

from migrant import exceptions
from migrant.backend import MigrantBackend
from migrant.repository import Repository


log = logging.getLogger(__name__)

Actions = List[Tuple[str, str]]

DBN = TypeVar("DBN")
DBC = TypeVar("DBC")


class MigrantEngine(Generic[DBN, DBC]):
    def __init__(
        self,
        backend: MigrantBackend[DBN, DBC],
        repository: Repository,
        config: Dict[str, str],
        dry_run: bool = False,
        processes: Optional[int] = None,
    ) -> None:
        self.backend = backend
        self.repository = repository
        self.script_ids = ["INITIAL"] + repository.list_script_ids()
        self.dry_run = dry_run
        self.config = config
        self.processes = processes or multiprocessing.cpu_count()

    def status(self, target_id: Optional[str] = None) -> int:
        """Return number of migration actions to be performed to
        upgrade to target_id"""
        target_id = self.pick_rev_id(target_id)
        conns = self.backend.generate_connections()

        total_actions = 0
        for db in conns:
            try:
                cdb = self.initialized_db(db)
            except exceptions.DatabaseUnavailable:
                continue
            actions = self.calc_actions(cdb, target_id)
            total_actions += len(actions)

        return total_actions

    def _update(self, db: DBN, target_id: str) -> None:
        try:
            cdb = self.initialized_db(db)
        except exceptions.DatabaseUnavailable:
            return
        log.info(f"{_pname()}: Starting migration for {cdb}")
        actions = self.calc_actions(cdb, target_id)
        try:
            self.execute_actions(cdb, actions)
            self.backend.commit(cdb)
        except:
            self.backend.abort(cdb)
            raise
        finally:
            self.backend.cleanup(cdb)
        log.info(f"{_pname()}: Migration completed for {cdb}")

    def update(self, target_id: Optional[str] = None) -> None:
        target_id = self.pick_rev_id(target_id)
        conns = self.backend.generate_connections()

        f = functools.partial(self._update, target_id=target_id)

        if self.processes == 1:
            for conn in conns:
                f(conn)
        else:
            with multiprocessing.Pool(self.processes) as pool:
                for _ in pool.imap_unordered(f, conns):
                    pass

    def test(self, target_id: Optional[str] = None) -> None:
        target_id = self.pick_rev_id(target_id)
        conns = self.backend.generate_test_connections()

        for db in conns:
            try:
                cdb = self.initialized_db(db)
            except exceptions.DatabaseUnavailable:
                continue
            actions = self.calc_actions(cdb, target_id)

            # Perform 2 passes of up/down to make sure database is still
            # upgradeable after being downgraded.
            for testpass in range(1, 3):
                log.info(f"PASS {testpass}. Testing upgrade for {cdb}")
                self.execute_actions(cdb, actions, strict=True)

                log.info(f"PASS {testpass}. Testing downgrade for {cdb}")
                reverted_actions = self.revert_actions(actions)
                self.execute_actions(cdb, reverted_actions, strict=True)

            log.info("Testing completed for %s" % cdb)

    def initialized_db(self, db: DBN) -> DBC:
        log.info(f"{_pname()}: Preparing migrations for {db}")
        try:
            cdb = self.backend.begin(db)
        except exceptions.DatabaseUnavailable:
            log.warning(f"{_pname()}: Starting migration for {db}")
            raise

        migrations = self.backend.list_migrations(cdb)
        if not migrations:
            latest_revid = self.pick_rev_id(None)
            self.initialize_db(cdb, latest_revid)

        return cdb

    def initialize_db(self, db: DBC, initial_revid: str):
        """Iniitialize database that was never migrated before

        Assume it is fully up-to-date.
        """
        # We can assuming the current database state is fully up-to-date. This
        # is the same thing as if all past migrations were executed.
        for sid in self.script_ids:
            if sid != "INITIAL":
                # Try to resolve into proper script name
                script = self.repository.load_script(sid)
                sid = script.name
            self.backend.push_migration(db, sid)

        log.info(
            f"{_pname()}: Initialized migrations for {db}. Assuming database is at {sid}"
        )

    def pick_rev_id(self, rev_id: Optional[str] = None) -> str:
        if rev_id is None:
            # Pick latest one
            rev_id = self.script_ids[-1]

        if canonical_rev_id(rev_id) not in self.script_ids:
            raise exceptions.ScriptNotFoundError(rev_id)

        return rev_id

    def calc_actions(self, db: DBC, target_revid: str) -> Actions:
        """Caclulate actions, required to update to revision `target_revid`
        """
        target_revid = canonical_rev_id(target_revid)
        assert target_revid in self.script_ids
        migrations = self.list_backend_migrations(db)
        assert len(migrations) > 0, "Migrations are initialized"

        script_idx = {v: idx for idx, v in enumerate(self.script_ids)}

        migrations = [m for m in migrations if m in script_idx]
        migrations = sorted(migrations, key=lambda m: script_idx[m])

        if not migrations:
            log.warning(
                "No common revision between repository and "
                "database %s. Running all migrations up to %s",
                db,
                target_revid,
            )
            base_revid = "INITIAL"
        else:
            base_revid = migrations[0]

        base_idx = script_idx[base_revid]
        target_idx = script_idx[target_revid]

        toremove = [m for m in reversed(migrations) if script_idx[m] > target_idx]
        toadd = [
            s
            for s in self.script_ids
            if base_idx < script_idx[s] <= target_idx and s not in migrations
        ]
        return [("-", rid) for rid in toremove] + [("+", rid) for rid in toadd]

    def revert_actions(self, actions: Actions) -> Actions:
        reverts = [("+" if a == "-" else "-", script) for a, script in actions]
        return list(reversed(reverts))

    def list_backend_migrations(self, db: DBC) -> List[str]:
        return [canonical_rev_id(revid) for revid in self.backend.list_migrations(db)]

    def execute_actions(self, db: DBC, actions: Actions, strict: bool = False) -> None:
        for action, revid in actions:
            script = self.repository.load_script(revid)
            assert action in ("+", "-")
            if action == "+":
                after = script.test_after_up
                before = script.test_before_up
                during = script.up
                end = self.backend.push_migration
                infinitive = "Upgrading"
            elif action == "-":
                after = script.test_after_down
                before = script.test_before_down
                during = script.down
                end = self.backend.pop_migration
                infinitive = "Reverting"
            log.info(
                "%s to %s%s",
                infinitive,
                script.name,
                " (not really)" if self.dry_run else "",
            )
            if not self.dry_run:
                if strict:
                    before(db)
                during(db)
                if strict:
                    after(db)
                end(db, script.name)


def _pname() -> str:
    return multiprocessing.current_process().name


def canonical_rev_id(migration_name: str) -> str:
    if "_" in migration_name:
        return migration_name.split("_", 1)[0]
    else:
        return migration_name



###############################################################################
#
# Copyright 2014 by Shoobx, Inc.
#
###############################################################################
from typing import List
import os
import logging
import string
import hashlib
import importlib.util

log = logging.getLogger(__name__)

from migrant import exceptions


INITIAL_SCRIPTLIST = """
# Order of migration scripts. This file is maintained by migrant
#
""".lstrip()


SCRIPT_TEMPLATE = '''"""%(title)s"""


def up(db):
    # Upgrade database
    pass


def down(db):
    # Downgrade database
    pass


# Tests for migration

def test_before_up(db):
    pass


def test_after_up(db):
    pass


def test_before_down(db):
    pass


def test_after_down(db):
    pass

'''.lstrip()


class Script:
    rev_id: str
    name: str

    def __init__(self, filename):
        assert filename.endswith(".py")
        self.name = os.path.basename(filename)[:-3]
        self.module = self._load_module_from_file(self.name, filename)

    def _load_module_from_file(self, name, path):
        spec = importlib.util.spec_from_file_location(name, path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    def up(self, db):
        self._exec("up", db)

    def down(self, db):
        self._exec("down", db)

    def test_before_up(self, db):
        self._exec("test_before_up", db)

    def test_after_up(self, db):
        self._exec("test_after_up", db)

    def test_before_down(self, db):
        self._exec("test_before_down", db)

    def test_after_down(self, db):
        self._exec("test_after_down", db)

    def _exec(self, method, *args, **kwargs):
        __traceback_info__ = (args, kwargs)
        if not hasattr(self.module, method):
            return

        return getattr(self.module, method)(*args, **kwargs)


class Repository:
    def new_script(self, title: str) -> str:
        raise NotImplementedError()

    def list_script_ids(self) -> List[str]:
        raise NotImplementedError()

    def load_script(self, scriptid: str) -> Script:
        raise NotImplementedError()


class DirectoryRepository(Repository):
    def __init__(self, directory):
        self.directory = directory
        self.scriptlist_fname = os.path.join(self.directory, "scripts.lst")

    def init(self):
        if not os.path.exists(self.directory):
            log.info("Creating migrations directory %s" % self.directory)
            os.makedirs(self.directory)

        if not os.path.exists(self.scriptlist_fname):
            log.info("Creating initial scripts.lst")
            with open(self.scriptlist_fname, "w") as f:
                f.write(INITIAL_SCRIPTLIST)

    def new_script(self, title):
        self.check_repo()

        # Make name out of title
        title = title.strip()
        toreplace = string.punctuation + " "
        trmap = str.maketrans(toreplace, "_" * len(toreplace))
        name = title.lower().translate(trmap)
        revid = hashlib.sha1((self.directory + title).encode("utf-8")).hexdigest()[:6]
        revname = f"{revid}_{name}"
        fname = "%s.py" % revname
        fullfname = os.path.join(self.directory, fname)

        if os.path.exists(fullfname):
            log.error("Script %s is already registered" % fname)
            raise exceptions.ScriptAlreadyExists()

        # Create script in repo
        with open(fullfname, "w") as sf:
            ns = {"title": title}
            sf.write(SCRIPT_TEMPLATE % ns)

        # Register script in list
        with open(self.scriptlist_fname, "a") as lf:
            lf.write(fname)
            lf.write("\n")

        log.info("Script %s created" % fullfname)
        return revname

    def list_script_ids(self):
        """List scripts in right order
        """
        self.check_repo()

        if not os.path.exists(self.scriptlist_fname):
            return []

        with open(self.scriptlist_fname, "r") as f:
            contents = f.readlines()

        scripts = []
        for scriptname in contents:
            scriptname = scriptname.strip()
            if scriptname.startswith("#"):
                continue

            if not scriptname:
                continue

            if not self.is_valid_scriptname(scriptname):
                log.warning("Ignoring unrecognized script name: %s" % scriptname)
                continue

            scripts.append(self.fname_to_revid(scriptname))

        return scripts

    def load_script(self, scriptid):
        self.check_repo()

        # Find script with given id
        fname = None
        for fname in os.listdir(self.directory):
            if not fname.endswith(".py"):
                continue

            if fname.startswith("%s_" % scriptid):
                break
        else:
            raise exceptions.ScriptNotFoundError(scriptid)

        return Script(os.path.join(self.directory, fname))

    def is_valid_scriptname(self, fname):
        return "_" in fname and fname.endswith(".py")

    def fname_to_revid(self, fname):
        return fname.split("_")[0]

    def check_repo(self):
        if not os.path.exists(self.directory):
            raise exceptions.RepositoryNotFound(self.directory)


def create_repo(cfg):
    return DirectoryRepository(cfg["repository"])



###############################################################################
#
# Copyright 2014 by Shoobx, Inc.
#
###############################################################################

import unittest
import tempfile
import os
import shutil
import textwrap

from migrant import repository


class RepositoryTest(unittest.TestCase):
    def setUp(self):
        self.dir = tempfile.mkdtemp("migrant")
        self.slistfname = os.path.join(self.dir, "scripts.lst")

    def tearDown(self):
        shutil.rmtree(self.dir)

    def test_create_list(self):
        repo = repository.DirectoryRepository(self.dir)
        repo.init()
        newrev = repo.new_script("Hello, World")

        revids = repo.list_script_ids()
        self.assertEqual(revids, [newrev.split("_")[0]])

    def test_is_valid_scriptname(self):
        repo = repository.DirectoryRepository(self.dir)
        self.assertTrue(repo.is_valid_scriptname("xxxx_hello.py"))
        self.assertFalse(repo.is_valid_scriptname("hello.py"))
        self.assertFalse(repo.is_valid_scriptname("# Some Comment"))
        self.assertFalse(repo.is_valid_scriptname("# Some_Comment"))

    def test_weird_list(self):
        repo = repository.DirectoryRepository(self.dir)
        repo.init()

        SCRIPTSLST = textwrap.dedent(
            """
        # This is some Comment

        a24bc_first.py
        weird line
        <<<< HEAD
        d724a_second.py
        ====
        1babe_third.py
        >>>> CONFLICT MARKER
        """
        ).lstrip()
        with open(self.slistfname, "w") as lf:
            lf.write(SCRIPTSLST)

        revids = repo.list_script_ids()
        self.assertEqual(revids, ["a24bc", "d724a", "1babe"])



###############################################################################
#
# Copyright 2014 by Shoobx, Inc.
#
###############################################################################
from typing import List, Iterable, Generic, TypeVar
import logging
import importlib.metadata

from migrant import exceptions

log = logging.getLogger(__name__)


# Database name type
DBN = TypeVar("DBN")

# Database connection type
DBC = TypeVar("DBC")


class MigrantBackend(Generic[DBN, DBC]):
    """Base interface for backend implementations"""

    def generate_connections(self) -> Iterable[DBN]:
        """Generate connections to process
        """
        raise NotImplementedError  # pragma: no cover

    def generate_test_connections(self) -> Iterable[DBN]:
        """Generate connections for migration tests

        The same interface as `generate_connections`, however this should not
        return "live" database connections, as tests might harm the data.

        If no connections are generated by this method, no tests will be run by
        `migrant test`.
        """
        return []  # pargma: no cover

    def begin(self, db: DBN) -> DBC:
        """Begin the migration

        This gives the opportunity to perform actual connection to the database
        before any migration is run.

        Can raise DatabaseUnavailable exception. In this case, migration will
        be skipped for this database.
        """
        raise NotImplementedError  # pragma: no cover

    def commit(self, db: DBC) -> None:
        """Called on successful completion of a migration

        This is an opportunity to commit work and close the connection.
        """
        pass

    def abort(self, db: DBC) -> None:
        """Called when migration have failed"""
        pass

    def cleanup(self, db: DBC) -> None:
        """Complete the migration

        Called when all work on database is done.
        """
        pass

    def list_migrations(self, db: DBC) -> List[str]:
        raise NotImplementedError  # pragma: no cover

    def push_migration(self, db: DBC, migration: str) -> None:
        raise NotImplementedError  # pragma: no cover

    def pop_migration(self, db: DBC, migration: str) -> None:
        raise NotImplementedError  # pragma: no cover

    def on_new_script(self, rev_name: str) -> None:
        """Called when new script is created
        """
        pass  # pragma: no cover

    def on_repo_init(self) -> None:
        """Called when new script repository is initialized
        """
        pass  # pragma: no cover


class NoopBackend(MigrantBackend[str, str]):
    def __init__(self, cfg):
        self.cfg = cfg

    def list_migrations(self, db: str) -> List[str]:
        return ["INITIAL"]

    def push_migration(self, db: str, migration: str) -> None:
        log.info("NOOP: pushing migration %s" % migration)

    def pop_migration(self, db: str, migration: str) -> None:
        log.info("NOOP: popping migration %s" % migration)

    def generate_connections(self):
        yield "NOOP"

    def begin(self, db: str) -> str:
        return db


def create_backend(cfg):
    name = cfg["backend"]

    factory = get_backend(name)
    return factory(cfg)


def get_backend(name):
    backends = importlib.metadata.entry_points(group="migrant").select(name=name)
    if not backends:
        raise exceptions.BackendNotRegistered(name)
    if len(backends) > 1:
        raise exceptions.BackendNameConflict(backends)
    pcls = backends[name].load()
    return pcls



###############################################################################
#
# Copyright 2014 by Shoobx, Inc.
#
###############################################################################

import logging

import pytest
import mock

@pytest.fixture
def migrant_backend():
    """Fixture to set up backend
    """

    class BackendSetter:
        def __init__(self, mocked_get_backend):
            self.mocked_get_backend = mocked_get_backend

        def set(self, backend):
            self.mocked_get_backend.return_value = lambda cfg: backend

    with mock.patch("migrant.backend.get_backend") as get_backend:
        yield BackendSetter(get_backend)


def pytest_configure(config):
    logging.root.addHandler(logging.NullHandler())
